#include "AdaptiveEnsembleStrategy.h"

#include <iostream>
#include <fstream>

#include "RandomGreedy.h"

using namespace std;

AdaptiveEnsembleStrategy::AdaptiveEnsembleStrategy(
	int d,
	function<double(double q, int k)> f,
	int k0,
	int l,
	double r,
	ClusteringInterface *final_algorithm,
	int k_max,
	bool iterative)
{
	d_ = d;
	f_ = f;
	k0_ = k0;
	l_ = l;
	r_ = r;
	final_algorithm_ = final_algorithm;
	k_max_ = k_max;
	iterative_ = iterative;
}

AdaptiveEnsembleStrategy::~AdaptiveEnsembleStrategy()
{
}

Partition *AdaptiveEnsembleStrategy::DetectCommunities(Graph *graph,
	Partition *partition,
	double &q)
{
	if (iterative_)
		return DetectCommunitiesIterative(graph, partition, q);

	vector<pair<double, Partition *>> clusterings;

	int k = k0_;

	int k_minus, k_plus;
	double q_minus, q_plus;
	double f_minus, f_plus;

	Partition *tmp_partition = nullptr;

	//sprintf_s(name, "%s%d%s", "Results/celegans/celegans", l_, ".csv");
	//ofstream fout(name);
	//fout << "step;k;q;f\n";

	for (int i = 0; i < l_; i++)
	{
		// Minus
		k_minus = max(1, k - d_);
		k_plus = min(k_max_, k + d_);

		tmp_partition = RandomGreedy(k_minus).DetectCommunities(graph,
			partition,
			q_minus);
		clusterings.push_back(make_pair(q_minus, tmp_partition));

		tmp_partition = RandomGreedy(k_plus).DetectCommunities(graph,
			partition,
			q_plus);
		clusterings.push_back(make_pair(q_plus, tmp_partition));

		f_minus = f_(q_minus, k_minus);
		f_plus = f_(q_plus, k_plus);

		/*cout << k << endl;
		cout << "-  " << k_minus << " " << q_minus << " " << f_minus << endl;
		cout << "+  " << k_plus << " " << q_plus << " " << f_plus << endl;*/
		//fout << i * 2 + 1 << ";" << k_minus << ";" << q_minus << ";" << f_minus << endl;
		//fout << i * 2 + 2 << ";" << k_plus << ";" << q_plus << ";" << f_plus << endl;

		k = round(max(1.0, k - (f_plus - f_minus) / (k_plus - k_minus)));
		k = min(k_max_, k);
	}

	sort(clusterings.begin(), clusterings.end(),
		[](pair<double, Partition *> &i, pair<double, Partition *> &j) -> bool
	{
		return i.first < j.first;
	}
	);

	//fout.close();
	int i = 0;
	//for (i = 0; i < 2 * l_ - l2; i++)
	while (clusterings[i].first <= (1.0 - r_) * clusterings.back().first)
	{
		delete clusterings[i].second;

		i++;
	}

	Partition *overlapping;
	overlapping = clusterings[i++].second;

	for ( ; i < 2 * l_; i++)
	{
		tmp_partition = Partition::OverlapClusters(overlapping,
			clusterings[i].second,
			graph->VertexCount());
		delete clusterings[i].second;
		delete overlapping;
		overlapping = tmp_partition;
	}

	Partition *result_partition = final_algorithm_->DetectCommunities(graph,
		overlapping,
		q);
	
	delete overlapping;
	return result_partition;
}

Partition *AdaptiveEnsembleStrategy::DetectCommunitiesIterative(Graph *graph,
	Partition *partition,
	double &q)
{
	Partition *previous_partition = partition;
	double prev_q = -1;

	//ofstream fout("Results/aes.csv");
	//int z = 0;
	//fout << "step;q\n";

	while (true)
	{
		vector<pair<double, Partition *>> clusterings;

		int k = k0_;

		int k_minus, k_plus;
		double q_minus, q_plus;
		double f_minus, f_plus;

		Partition *tmp_partition = nullptr;

		for (int i = 0; i < l_; i++)
		{
			k_minus = max(1, k - d_);
			k_plus = min(k_max_, k + d_);

			tmp_partition = RandomGreedy(k_minus).DetectCommunities(graph,
				previous_partition,
				q_minus);
			clusterings.push_back(make_pair(q_minus, tmp_partition));

			tmp_partition = RandomGreedy(k_plus).DetectCommunities(graph,
				previous_partition,
				q_plus);
			clusterings.push_back(make_pair(q_plus, tmp_partition));

			f_minus = f_(q_minus, k_minus);
			f_plus = f_(q_plus, k_plus);

			k = round(max(1.0, k - (f_plus - f_minus) / (k_plus - k_minus)));
			k = min(k_max_, k);
		}

		sort(clusterings.begin(), clusterings.end(),
			[](pair<double, Partition *> &i, pair<double, Partition *> &j) -> bool
		{
			return i.first < j.first;
		}
		);

		int l2 = ceil(2 * r_ * l_);

		int i = 0;
		//for (i = 0; i < 2 * l_ - l2; i++)
		while (clusterings[i].first <= (1 - r_) * clusterings.back().first)
		{
			//cout << clusterings[i].first << " ";
			delete clusterings[i].second;			
			i++;
		}

		//cout << endl << "---" << i << endl;

		Partition *overlapping;
		overlapping = clusterings[i++].second;

		for (; i < 2 * l_; i++)
		{
			//cout << clusterings[i].first << " ";
			tmp_partition = Partition::OverlapClusters(overlapping,
				clusterings[i].second,
				graph->VertexCount());
			delete clusterings[i].second;
			delete overlapping;
			overlapping = tmp_partition;
		}
		//cout << endl;

		double overlapping_q = overlapping->Modularity(graph);

		//cout << endl << "===" <<  prev_q << " " << overlapping_q << endl;
		//fout << z++ << ";" << overlapping_q << endl;
		if (overlapping_q > prev_q)
		{
			if (previous_partition != partition)
				delete previous_partition;

			previous_partition = overlapping;
			prev_q = overlapping_q;
			continue;
		}

		delete overlapping;

		Partition *result_partition = final_algorithm_->DetectCommunities(graph,
			previous_partition,
			q);

		//fout.close();

		delete previous_partition;
		return result_partition;
	}
}

